from django_bundles.tests.utils import *
from django_bundles.tests.utils.files import *
from django_bundles.tests.utils.processes import *

from django_bundles.tests.conf import *

from django_bundles.tests.core import *

from django_bundles.tests.templatetags import *


import subprocess
import errno
import select
import os


_PIPE_BUF = getattr(select, 'PIPE_BUF', 512)


def run_process(cmd, stdin=None, iterate_stdin=True, output_chunk_size=1024, shell=True, to_close=None):
    """
    This is a modification of subprocess.Popen.communicate that accepts an iterable stdin and is itself a generator for stdout
    """
    try:
        p = subprocess.Popen(cmd, shell=shell, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        if stdin:
            if iterate_stdin:
                stdin_iter = iter(stdin)
                stdin_buffer = ''
                stdin_available = True
            else:
                stdin_buffer = stdin
                stdin_available = False

        write_set = []
        read_set = []
        output_buffer = ''

        if p.stdin and stdin:
            write_set.append(p.stdin)
        if p.stdout:
            read_set.append(p.stdout)
        if p.stderr:
            read_set.append(p.stderr)

        while read_set or write_set:
            try:
                rlist, wlist, xlist = select.select(read_set, write_set, [])
            except select.error as e:
                if e.args[0] == errno.EINTR:
                    continue
                raise

            if p.stdin in wlist:
                while len(stdin_buffer) < _PIPE_BUF and stdin_available:
                    try:
                        stdin_buffer += stdin_iter.next()
                    except StopIteration:
                        stdin_available = False
                chunk = stdin_buffer[:_PIPE_BUF]
                bytes_written = os.write(p.stdin.fileno(), chunk)
                stdin_buffer = stdin_buffer[bytes_written:]

                if not (stdin_buffer or stdin_available):
                    p.stdin.close()
                    write_set.remove(p.stdin)

            if p.stdout in rlist:
                data = os.read(p.stdout.fileno(), output_chunk_size)
                if data == '':
                    p.stdout.close()
                    read_set.remove(p.stdout)
                if data:
                    output_buffer += data
                    yield data

            if p.stderr in rlist:
                data = os.read(p.stderr.fileno(), output_chunk_size)
                if data == '':
                    p.stderr.close()
                    read_set.remove(p.stderr)
                if data:
                    output_buffer += data

            if len(output_buffer) > output_chunk_size:
                output_buffer = output_buffer[-output_chunk_size:]

        return_code = p.poll()
        if return_code:
            e = subprocess.CalledProcessError(return_code, cmd)
            e.output = output_buffer
            raise e
    finally:
        if to_close:
            to_close.close()



from django.core.management.base import BaseCommand

from django_bundles.conf.bundles_settings import bundles_settings
from django_bundles.core import get_bundles, set_bundle_versions
from django_bundles.processors import processor_pipeline, processor_library
from django_bundles.utils.files import FileChunkGenerator

import os
from hashlib import md5


def iter_bundle_files(bundle, debug=False):
    for bundle_file in bundle.files:
        for chunk in processor_pipeline(bundle_file.processors, FileChunkGenerator(open(bundle_file.file_path, 'rb')), debug=debug):
            yield chunk
        yield '\n'


def make_bundle(bundle, debug=False):
    """
    Does all of the processing required to create a bundle and write it to disk, returning its hash version
    """
    iter_input = iter_bundle_files(bundle, debug=debug)

    output_pipeline = processor_pipeline(bundle.processors, iter_input, debug=debug)

    m = md5()

    tmp_output_file_name = '%s.%s.%s' % (os.path.join(bundle.bundle_file_root, bundle.bundle_filename), 'temp', bundle.bundle_type)

    with open(tmp_output_file_name, 'wb') as output_file:
        for chunk in output_pipeline:
            m.update(chunk)
            output_file.write(chunk)

    hash_version = m.hexdigest()

    if debug:
        output_file_name = '%s.debug.%s.%s' % (os.path.join(bundle.bundle_file_root, bundle.bundle_filename), hash_version, bundle.bundle_type)
    else:
        output_file_name = '%s.%s.%s' % (os.path.join(bundle.bundle_file_root, bundle.bundle_filename), hash_version, bundle.bundle_type)

    os.rename(tmp_output_file_name, output_file_name)

    return hash_version


class Command(BaseCommand):
    help = "Bundles up the media"
    requires_model_validation = False

    def handle(self, *args, **options):
        self.stdout.write("Bundling...\n")

        _bundle_versions = {}
        set_bundle_versions(_bundle_versions)

        for bundle in get_bundles():
            self.stdout.write("Writing bundle: %s\n" % bundle.name)

            hash_version = make_bundle(bundle)

            # Build bundle versions as we're going along in case they're used in templated bundles
            _bundle_versions[bundle.name] = hash_version

            if bundle.create_debug:
                _bundle_versions['debug:' + bundle.name] = make_bundle(bundle, debug=True)

            self.stdout.write("\t%s\n" % bundle.get_version())

        version_info = '\n'.join(['    "%s": "%s",' % version for version in _bundle_versions.iteritems()])

        with open(bundles_settings.BUNDLES_VERSION_FILE, 'wb') as bundles_versions:
            bundles_versions.write("""\
#!/usr/bin/env python

BUNDLES_VERSIONS = {
%s
}
""" % version_info)

        for single_file_input, single_file_output in bundles_settings.BUNDLES_SINGLE_FILES:
            self.stdout.write("Writing: %s\n" % single_file_output)
            file_type = os.path.splitext(single_file_input)[1][1:]
            processors = processor_library.get_default_preprocessors_for(file_type) + processor_library.get_default_postprocessors_for(file_type)

            with open(single_file_output, 'wb') as output_file:
                for chunk in processor_pipeline(processors, FileChunkGenerator(open(single_file_input, 'rb'))):
                    output_file.write(chunk)

        self.stdout.write("Done.\n")



from subprocess import CalledProcessError
from django.core.management.base import BaseCommand, CommandError

from django_bundles.core import get_bundles
from django_bundles.processors import processor_pipeline
from django_bundles.utils.files import FileChunkGenerator
from django_bundles.utils.processes import run_process
from django_bundles.conf.bundles_settings import bundles_settings

from optparse import make_option
import collections

import os
from tempfile import NamedTemporaryFile


class Command(BaseCommand):
    help = "Lints the bundles based on settings.BUNDLES_LINTING"
    option_list = BaseCommand.option_list + (
        make_option('--failures-only',
            action='store_true',
            default=False,
            help='Only report failures'
        ),
        make_option('--pattern',
            help='Simple pattern matching for files',
        ),
        make_option('--watch',
            action='store_true',
            default=False,
            help='Watch files for changes'
        ),
    )
    requires_model_validation = False

    def lint_file(self, bundle_type, file_path, iter_input=None):
        command = bundles_settings.BUNDLES_LINTING[bundle_type]['command']

        input_file = None
        stdin = None

        if '{infile}' in command:
            if iter_input:
                if hasattr(iter_input, 'file_path'):
                    filename = iter_input.file_path
                else:
                    input_file = NamedTemporaryFile()
                    for chunk in iter_input:
                        input_file.write(chunk)

                    filename = input_file.name
            else:
                filename = file_path

            command = command.format(infile=filename)
        else:
            if iter_input:
                stdin = iter_input
            else:
                stdin = input_file = open(file_path, 'rb')

        try:
            # Consume the iterator into a zero length deque
            collections.deque(run_process(command, stdin=stdin, to_close=input_file), maxlen=0)
        except CalledProcessError as e:
            return False, e.output

        return True, ''


    def drawscreen(self):
        import curses

        self.stdout.write(curses.tigetstr('clear'))
        self.stdout.write("Watching files for changes...")

        if self.errored_files:
            self.stdout.write(self.style.ERROR('%s file%s to fix' % (len(self.errored_files), 's' if len(self.errored_files) > 1 else '')))
        else:
            self.stdout.write(self.style.HTTP_REDIRECT('no files to fix'))

        self.stdout.write("\n")

        self.stdout.write(curses.tigetstr('cud1') * 2)

        for filename, error_message in self.errored_files.iteritems():
            self.stdout.write('\t' + self.style.HTTP_SERVER_ERROR(filename) + '\n')
            self.stdout.write(self.style.HTTP_SERVER_ERROR('\n'.join(['\t\t' + error_line for error_line in error_message.split('\n')])) + '\n\n')

        self.stdout.write(curses.tigetstr('cud1') * 2)

        for log_line in reversed(self.log_lines):
            self.stdout.write('\t' + log_line + '\n')


    def log_watch_result(self, src, result, error_message=None):
        if result:
            if src in self.errored_files:
                del self.errored_files[src]
            self.log_lines.append(self.style.HTTP_SUCCESS('OK\t\t%s' % src))
        else:
            self.log_lines.append(self.style.HTTP_SERVER_ERROR('FAIL\t\t%s' % src))
            self.errored_files[src] = self.style.HTTP_SERVER_ERROR(error_message)

        self.log_lines = self.log_lines[-5:]
        self.drawscreen()


    def handle(self, *args, **options):
        self.show_successes = not bool(options.get('failures_only'))
        watch = bool(options.get('watch'))
        file_pattern = options.get('pattern')


        if watch:
            try:
                import time
                from watchdog.observers import Observer
                from watchdog.events import FileSystemEventHandler
                import curses
            except ImportError:
                raise CommandError('watchdog is required for this (pip install watchdog')

            self.errored_files = {}
            self.log_lines = []
            watching = {}

            def check_and_lint_file(src):
                # TODO: don't repeatedly lint the same file
                for watchdir in watching:
                    if watchdir in src:
                        for bundle in watching[watchdir]:
                            if src in bundle:
                                result, error_message = self.lint_file(bundle.bundle_type, src, iter_input=processor_pipeline(bundle[src].processors, FileChunkGenerator(open(src, 'rb'))))
                                self.log_watch_result(src, result, error_message=error_message)
                                break

            class FileEventHandler(FileSystemEventHandler):
                def on_created(self, event):
                    if not event.is_directory:
                        check_and_lint_file(event.src_path)

                def on_modified(self, event):
                    if not event.is_directory:
                        check_and_lint_file(event.src_path)

            # TODO: watchdog dirsnapshot line 97 patched (otherwise it doesn't work with PyCharm)
            #        #if stat_info.st_ino == ref_stat_info.st_ino and stat_info.st_mtime != ref_stat_info.st_mtime:
            #        if stat_info.st_mtime != ref_stat_info.st_mtime:

            event_handler = FileEventHandler()
            observer = Observer()
            curses.setupterm()
            self.drawscreen()

            for bundle in get_bundles():
                if bundle.files_root not in watching:
                    watching[bundle.files_root] = set()
                    observer.schedule(event_handler, path=bundle.files_root, recursive=True)
                watching[bundle.files_root].add(bundle)

            observer.start()
            try:
                while True:
                    time.sleep(10)
            except KeyboardInterrupt:
                observer.stop()
            observer.join()

            return


        files_linted = set()

        failures = 0

        def file_checked(success, error_message, file_path):
            files_linted.add(file_path)

            if success:
                if self.show_successes:
                    self.stdout.write(self.style.HTTP_SUCCESS('OK\t\t%s\n' % file_path))
                return 0
            else:
                self.stdout.write(self.style.HTTP_SERVER_ERROR('FAIL\t\t%s\n' % file_path))
                self.stdout.write(self.style.HTTP_SERVER_ERROR(error_message))
                return 1


        for bundle in get_bundles():
            for bundle_file in bundle.files:
                if file_pattern and file_pattern not in bundle_file.file_path:
                    continue
                if bundle_file.file_path in files_linted:
                    continue

                # Check the file exists, even for non-linted files
                if not os.path.exists(bundle_file.file_path):
                    self.stdout.write(self.style.HTTP_SERVER_ERROR('FAIL\t\t%s\n' % bundle_file.file_path))
                    self.stdout.write(self.style.HTTP_SERVER_ERROR('File does not exist (referenced from %s)\n' % bundle.name))
                    failures += 1
                    continue

                if not bundle_file.lint:
                    continue

                success, error_message = self.lint_file(bundle.bundle_type, bundle_file.file_path, iter_input=processor_pipeline(bundle_file.processors, FileChunkGenerator(open(bundle_file.file_path, 'rb'))))

                failures += file_checked(success, error_message, bundle_file.file_path)


        for single_file_path, _ in bundles_settings.BUNDLES_SINGLE_FILES:
            success, error_message = self.lint_file(os.path.splitext(single_file_path)[1][1:], single_file_path)
            failures += file_checked(success, error_message, single_file_path)


        if failures:
            raise CommandError('%s FILE%s FAILED' % (failures, 'S' if failures > 1 else ''))
        else:
            self.stdout.write(self.style.HTTP_REDIRECT('\nALL FILES PASSED\n'))



from django.test import TestCase


from django_bundles.utils.files import expand_file_names, FileChunkGenerator


import os
import collections


class ExpandFileNamesTest(TestCase):
    def test_expand_file_names(self):
        current_dir = os.path.dirname(__file__)
        files_in_dir = os.listdir(current_dir)
        python_files_in_dir = [fn for fn in files_in_dir if fn[-3:] == '.py']

        expanded_file_list = expand_file_names("*.py", current_dir)

        self.assertEqual(expanded_file_list, python_files_in_dir)


class FileChunkGeneratorTest(TestCase):
    def test_contents(self):
        with open(__file__, 'r') as f:
            file_contents = f.read()

        with open(__file__, 'r') as f:
            g = FileChunkGenerator(f)

            file_contents_2 = ''
            for chunk in g:
                file_contents_2 += chunk

        self.assertEqual(file_contents, file_contents_2)

    def test_file_closes(self):
        with open(__file__, 'r') as f:
            g = FileChunkGenerator(f)

            collections.deque(g, maxlen=0)

            self.assertTrue(f.closed)

    def test_file_stays_open(self):
        with open(__file__, 'r') as f:
            g = FileChunkGenerator(f, close=False)

            collections.deque(g, maxlen=0)

            self.assertFalse(f.closed)



import mimetypes
import posixpath
import urllib
import os
from operator import concat
from django.http import HttpResponse
from django.views.static import serve as django_serve

from django_bundles.core import get_bundles
from django_bundles.processors import processor_pipeline
from django_bundles.utils.files import FileChunkGenerator

from django.conf import settings


file_cache = {}


def get_file(path):
    global file_cache

    if not file_cache:
        for bundle in get_bundles():
            for bundle_file in bundle.files:
                file_cache[os.path.realpath(bundle_file.file_path)] = {
                    'bundle_file': bundle_file,
                    'cache': None
                }

    if path in file_cache:
        if not file_cache[path]['cache']:
            mimetype, encoding = mimetypes.guess_type(path)
            mimetype = mimetype or 'application/octet-stream'
            # TODO: less files need to change the way they are rendered in the template
            print "Generating", path

            file_cache[path]['cache'] = {
                'contents': reduce(concat, (chunk for chunk in processor_pipeline(file_cache[path]['bundle_file'].processors, FileChunkGenerator(open(file_cache[path]['bundle_file'].file_path, 'rb'))))),
                'mimetype': mimetype,
            }

        return file_cache[path]['cache']

    return None



def serve(request, path, document_root=None, show_indexes=False):
    if not settings.USE_BUNDLES:
        path = posixpath.normpath(urllib.unquote(path))
        path = path.lstrip('/')
        newpath = ''
        for part in path.split('/'):
            if not part:
                # Strip empty path components.
                continue
            drive, part = os.path.splitdrive(part)
            head, part = os.path.split(part)
            if part in (os.curdir, os.pardir):
                # Strip '.' and '..' in path.
                continue
            newpath = os.path.join(newpath, part).replace('\\', '/')

        fullpath = os.path.join(document_root, newpath)

        cached = get_file(fullpath)
        if cached:
            return HttpResponse(cached['contents'], mimetype=cached['mimetype'])

    return django_serve(request, path, document_root=document_root, show_indexes=show_indexes)


import os

from django.core.management.base import BaseCommand

from django_bundles.core import get_bundles


class Command(BaseCommand):
    args = "target_directory"
    help = "Writes out files containing the list of input files for each bundle"
    requires_model_validation = False

    def handle(self, target_directory, *args, **options):
        try:
            os.mkdir(target_directory)
        except OSError:
            pass

        for bundle in get_bundles():
            manifest_filename = os.path.join(target_directory, bundle.name) + '.manifest'
            with open(manifest_filename, 'w') as manifest:
                manifest.write("\n".join(f.file_path for f in bundle.files))



from django.test import TestCase


from django_bundles.utils.processes import run_process


import collections


class RunProcessTest(TestCase):

    def test_command_runs(self):
        stdin = 'TEST'
        output = ''.join(run_process('cat', stdin=stdin, iterate_stdin=False))

        self.assertEqual(stdin, output)


    def test_piped_command(self):
        stdin = "TEST1\nTEST2\nTEST1"

        output = ''.join(run_process('cat | grep "TEST1"', stdin=stdin, iterate_stdin=False))

        self.assertEqual(output, "TEST1\nTEST1\n")


    def test_stdin_iter(self):
        stdin_iter = ("TEST%s" % x for x in xrange(0, 3))

        output = ''.join(run_process('cat', stdin=stdin_iter))

        self.assertEqual(output, "TEST0TEST1TEST2")


    def test_to_close(self):
        with open(__file__, 'r') as f:
            collections.deque(run_process('cat', stdin='TEST', iterate_stdin=False, to_close=f), maxlen=0)
            self.assertTrue(f.closed)



from django.test import TestCase
from django.conf import settings


from django_bundles.core import Bundle
from django_bundles.processors import ExecutableProcessor


import os


class BundleConfTest(TestCase):
    def setUp(self):
        self.bundle = Bundle((
            'test_bundle', {
                'type': 'css',
                'files': (
                    'test*.css',
                    ('another.css', {
                        'processors': (
                            ('django_bundles.processors.ExecutableProcessor', {'command': 'cat {infile}'}),
                        ),
                    }),
                ),
                'media': 'screen',
                'files_root': os.path.join(os.path.dirname(__file__), 'files'),
                'processors': (
                    ('django_bundles.processors.ExecutableProcessor', {'command': 'cat {infile}'}),
                ),
                'debug': True,
            }
        ))


    def test_basic_conf(self):
        self.assertEqual(self.bundle.name, 'test_bundle')
        self.assertEqual(self.bundle.files_root, os.path.join(os.path.dirname(__file__), 'files'))
        self.assertEqual(self.bundle.media, 'screen')
        self.assertEqual(self.bundle.bundle_type, 'css')
        self.assertEqual(self.bundle.bundle_filename, 'test_bundle')
        self.assertEqual(self.bundle.bundle_url_root, settings.MEDIA_URL)
        self.assertEqual(self.bundle.bundle_file_root, os.path.join(os.path.dirname(__file__), 'files'))
        self.assertTrue(self.bundle.create_debug)


    def test_processor_conf(self):
        self.assertEqual(len(self.bundle.processors), 1)
        self.assertTrue(isinstance(self.bundle.processors[0], ExecutableProcessor))
        self.assertEqual(self.bundle.processors[0].command, 'cat {infile}')


    def test_file_conf(self):
        self.assertEqual(len(self.bundle.files), 3)
        self.assertEqual([f.file_path for f in self.bundle.files], [
            os.path.join(os.path.dirname(__file__), 'files/test1.css'),
            os.path.join(os.path.dirname(__file__), 'files/test2.css'),
            os.path.join(os.path.dirname(__file__), 'files/another.css'),
        ])
        self.assertEqual([f.file_url for f in self.bundle.files], [
            os.path.join(settings.MEDIA_URL, 'test1.css'),
            os.path.join(settings.MEDIA_URL, 'test2.css'),
            os.path.join(settings.MEDIA_URL, 'another.css'),
        ])



from django.views.debug import get_safe_settings
from django.template import Context, Template

from django_bundles.core import get_bundles
from django_bundles.processors.base import Processor


class DjangoTemplateProcessor(Processor):
    """
    Processor that runs Django's templating code across the file - 'settings' and 'bundles' are passed into context
    Override this class to add other things to context
    """
    def get_context(self):
        """
        Override this to change context
        """
        return {
            'settings': get_safe_settings(),
            'bundles': get_bundles(),
            }

    def process(self, iter_input):
        template = Template(''.join(iter_input))
        yield template.render(Context(self.get_context()))




import os
from django.core.management.base import BaseCommand

from django_bundles.conf.bundles_settings import bundles_settings
from django_bundles.core import get_bundles, get_bundle_versions

class Command(BaseCommand):
    help = "Removes any created bundles"
    requires_model_validation = False

    def handle(self, *args, **options):

        bundle_versions = get_bundle_versions()

        for bundle in get_bundles():
            hash_version = bundle_versions[bundle.name]
            bundle_path = '%s.%s.%s' % (os.path.join(bundle.bundle_file_root, bundle.bundle_filename), hash_version, bundle.bundle_type)
            self.stdout.write("Removing bundle: %s\n" % bundle_path)
            try:
                os.remove(bundle_path)
            except:
                self.stderr.write("Could not remove bundle: %s\n" % bundle_path)

            if bundle.create_debug:
                debug_hash_version = bundle_versions['debug:' + bundle.name]

                bundle_debug_path = '%s.debug.%s.%s' % (os.path.join(bundle.bundle_file_root, bundle.bundle_filename), debug_hash_version, bundle.bundle_type)
                try:
                    os.remove(bundle_debug_path)
                except:
                    self.stderr.write("Could not remove debug bundle: %s\n" % bundle_debug_path)

        self.stdout.write("Removing bundles version file: %s\n" % bundles_settings.BUNDLES_VERSION_FILE)
        os.remove(bundles_settings.BUNDLES_VERSION_FILE)

        for _, single_file_output in bundles_settings.BUNDLES_SINGLE_FILES:
            self.stdout.write("Removing: %s\n" % single_file_output)
            try:
                os.remove(single_file_output)
            except:
                self.stderr.write("Could not remove single file: %s\n" % single_file_output)

        self.stdout.write("Done.\n")



def get_class(class_string):
    """
    Get a class from a dotted string
    """
    split_string = class_string.encode('ascii').split('.')
    import_path = '.'.join(split_string[:-1])
    class_name = split_string[-1]

    if class_name:
        try:
            if import_path:
                mod = __import__(import_path, globals(), {}, [class_name])
                cls = getattr(mod, class_name)
            else:
                cls = __import__(class_name, globals(), {})
            if cls:
                return cls
        except (ImportError, AttributeError):
            pass

    return None



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from django_bundles.conf import default_settings


class SettingsHelper(object):
    def __init__(self, user_settings, default_settings, optional_defaults):
        self.user_settings = user_settings
        self.default_settings = default_settings
        self.optional_defaults = set(optional_defaults)


    def __getattr__(self, name):
        if hasattr(self.user_settings, name):
            return getattr(self.user_settings, name)

        if name in self.optional_defaults:
            return getattr(self.default_settings, name)

        raise ImproperlyConfigured, "%s is a required setting for django_bundles" % name


bundles_settings = SettingsHelper(settings, default_settings, [
    'USE_BUNDLES',
    'DEFAULT_PREPROCESSORS',
    'DEFAULT_POSTPROCESSORS',
    'BUNDLES_LINTING',
    'BUNDLES_LINT_SUCCESS_OK',
    'BUNDLES_SINGLE_FILES',
    'BUNDLES_TAG_HTML',
])



from distutils.core import setup
from setuptools import find_packages

setup(name='django-bundles',
      version='0.3.5',
      description='Another Django media bundler',
      author='Sam Cooke',
      author_email='sam.cooke@xylate.com',
      url='https://github.com/sdcooke/django_bundles',
      packages=find_packages(),
      include_package_data=True,
      license="MIT license, see LICENSE file",
)
